##########################################################################
## FESSA COLOR PALETTE
##########################################################################
#  https://github.com/luigibonati/fessa-color-palette/blob/master/fessa.py

import mlcolvar
from matplotlib.colors import LinearSegmentedColormap, ColorConverter
from matplotlib.cm import register_cmap

__all__ = ["paletteFessa", "paletteCortina"]

# Fessa colormap
paletteFessa = [
    "#1F3B73",  # dark-blue
    "#2F9294",  # green-blue
    "#50B28D",  # green
    "#A7D655",  # pisello
    "#FFE03E",  # yellow
    "#FFA955",  # orange
    "#D6573B",  # red
]

cm_fessa = LinearSegmentedColormap.from_list("fessa", paletteFessa)
register_cmap(cmap=cm_fessa)
register_cmap(cmap=cm_fessa.reversed())

for i in range(len(paletteFessa)):
    ColorConverter.colors[f"fessa{i}"] = paletteFessa[i]

### To set it as default
# import fessa
# plt.set_cmap('fessa')
### or the reversed one
# plt.set_cmap('fessa_r')
### For contour plots
# plt.contourf(X, Y, Z, cmap='fessa')
### For standard plots
# plt.plot(x, y, color='fessa0')


# Cortina1980 colormap
paletteCortina = [
    [0.0, 0.0, 0.803921568627451, 1],  # mediumblue
    [0.4823529411764706, 0.40784313725490196, 0.9333333333333333, 1],  # mediumslateblue
    [0.0, 0.9803921568627451, 0.6039215686274509, 1],  # mediumspringgreen
    [0.23529411764705882, 0.7019607843137254, 0.44313725490196076, 1],  # mediumseagreen
    [0.8588235294117647, 0.4392156862745098, 0.5764705882352941, 1],  # palevioletred
    [0.7803921568627451, 0.08235294117647059, 0.5215686274509804, 1],  # mediumvioletred
]

cm_cortina = LinearSegmentedColormap.from_list("cortina80", paletteCortina)
register_cmap(cmap=cm_cortina)
register_cmap(cmap=cm_cortina.reversed())

##########################################################################
## HELPER FUNCTIONS FOR 2D SYSTEMS
##########################################################################

import matplotlib.pyplot as plt
import numpy as np
import torch


def muller_brown_potential(x, y):
    """Muller-Brown analytical potential"""
    prefactor = 0.15
    A = (-200, -100, -170, 15)
    a = (-1, -1, -6.5, 0.7)
    b = (0, 0, 11, 0.6)
    c = (-10, -10, -6.5, 0.7)
    x0 = (1, 0, -0.5, -1)
    y0 = (0, 0.5, 1.5, 1)
    offset = -146.7

    v = -prefactor * offset
    for i in range(4):
        v += (
            prefactor
            * A[i]
            * np.exp(
                a[i] * (x - x0[i]) ** 2
                + b[i] * (x - x0[i]) * (y - y0[i])
                + c[i] * (y - y0[i]) ** 2
            )
        )
    return v


def muller_brown_mfep():
    mfep = np.loadtxt(
        mlcolvar.__path__[0]
        + "/../docs/notebooks/tutorials/data/muller-brown/mfep.txt",
        usecols=(0, 1),
    )
    return mfep


def muller_brown_potential_three_states(x, y):
    """Muller-Brown analytical potential"""
    prefactor = 0.15
    A = (-280, -170, -170, 15)
    a = (-15, -1, -6.5, 0.7)
    b = (0, 0, 11, 0.6)
    c = (-10, -10, -6.5, 0.7)
    x0 = (1, 0.2, -0.5, -1)
    y0 = (0, 0.5, 1.5, 1)
    offset = -146.7

    v = -prefactor * offset
    for i in range(4):
        v += (
            prefactor
            * A[i]
            * np.exp(
                a[i] * (x - x0[i]) ** 2
                + b[i] * (x - x0[i]) * (y - y0[i])
                + c[i] * (y - y0[i]) ** 2
            )
        )
    return v


def muller_brown_three_states_mfep():
    mfep = np.loadtxt(
        mlcolvar.__path__[0]
        + "/../docs/notebooks/tutorials/data/muller-brown-3states/mfep.txt",
        usecols=(0, 1),
    )
    return mfep


def plot_isolines_2D(
    function,
    component=None,
    limits=((-1.8, 1.2), (-0.4, 2.1)),
    num_points=(100, 100),
    mode="contourf",
    levels=12,
    cmap=None,
    colorbar=None,
    max_value=None,
    ax=None,
    **kwargs,
):
    """Plot isolines of a function/model in a 2D space."""

    # Define grid where to evaluate function
    if type(num_points) == int:
        num_points = (num_points, num_points)
    xx = np.linspace(limits[0][0], limits[0][1], num_points[0])
    yy = np.linspace(limits[1][0], limits[1][1], num_points[1])
    xv, yv = np.meshgrid(xx, yy)

    # if torch module
    if isinstance(function, torch.nn.Module):
        z = np.zeros_like(xv)
        for i in range(num_points[0]):
            for j in range(num_points[1]):
                xy = torch.Tensor([xv[i, j], yv[i, j]])
                with torch.no_grad():
                    train_mode = function.training
                    function.eval()
                    s = function(xy).numpy()
                    function.training = train_mode
                    if component is not None:
                        s = s[component]
                z[i, j] = s
    # else apply function directly to grid points
    else:
        z = function(xv, yv)

    if max_value is not None:
        z[z > max_value] = max_value

    # Setup plot
    return_axs = False
    if ax is None:
        return_axs = True
        _, ax = plt.subplots(figsize=(6, 4.0), dpi=100)

    # Color scheme
    if cmap is None:
        if mode == "contourf":
            cmap = "fessa"
        elif mode == "contour":
            cmap = "Greys_r"

    # Colorbar
    if colorbar is None:
        if mode == "contourf":
            colorbar = True
        elif mode == "contour":
            colorbar = False

    # Plot
    if mode == "contourf":
        pp = ax.contourf(xv, yv, z, levels=levels, cmap=cmap, **kwargs)
        if colorbar:
            plt.colorbar(pp, ax=ax)
    else:
        pp = ax.contour(xv, yv, z, levels=levels, cmap=cmap, **kwargs)

    if return_axs:
        return ax
    else:
        return None


def plot_metrics(
    metrics,
    keys=["train_loss_epoch", "valid_loss"],
    x=None,  # 'epoch'
    labels=None,  # ['Train','Valid'],
    linestyles=None,  # ['-','--']
    colors=None,  # ['fessa0','fessa1']
    xlabel="Epoch",
    ylabel="Loss",
    title="Learning curves",
    yscale=None,
    ax=None,
):
    """Plot logged metrics."""

    # Setup axis
    return_axs = False
    if ax is None:
        return_axs = True
        _, ax = plt.subplots(figsize=(5, 4), dpi=100)

    # Plot metrics
    auto_x = True if x is None else False
    for i, key in enumerate(keys):
        y = metrics[key]
        lstyle = linestyles[i] if linestyles is not None else None
        label = labels[i] if labels is not None else key
        color = colors[i] if colors is not None else None
        if auto_x:
            x = np.arange(len(y))
        ax.plot(x, y, linestyle=lstyle, label=label, color=color)

    # Plot settings
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)
    if yscale is not None:
        ax.set_yscale(yscale)

    ax.legend(ncol=1, frameon=False)

    if return_axs:
        return ax
    else:
        return None


def test_utils_plot():
    import matplotlib

    x = np.linspace(-1.5, 1.5)
    y = np.linspace(-0.5, 2.5)

    mp = muller_brown_potential(x, y)
    mp = muller_brown_potential_three_states(x, y)

    pal = paletteFessa
    pal = paletteCortina

    cmap = matplotlib.colors.Colormap("fessa", 2)
    cmap = matplotlib.colors.Colormap("fessa_r", 2)
    cmap = matplotlib.colors.Colormap("cortina80", 2)
    cmap = matplotlib.colors.Colormap("cortina80_r", 2)



import torch

__all__ = ["Transform"]


class Transform(torch.nn.Module):
    """
    Base transform class.
    To implement a new transform override the forward method.
    The parameters of the transform should be set either in the initialization or via the setup_from_datamodule function.
    """

    def __init__(self, in_features: int, out_features: int):
        """Transform class options.

        Parameters
        ----------
        in_features : int
            Number of inputs of the transform
        out_features : int
            Number of outputs of the transform
        """
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features

    def setup_from_datamodule(self, datamodule):
        """
        Initialize parameters based on pytorch lighting datamodule.
        """
        pass

    def forward(self, X: torch.Tensor):
        raise NotImplementedError

    def teardown(self):
        pass



#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Tests for the members of the mlcolvar.data.datamodule module.
"""


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

import pytest
import torch

from mlcolvar.data.dataset import DictDataset
from mlcolvar.data.datamodule import DictModule


# =============================================================================
# TESTS
# =============================================================================


@pytest.mark.parametrize("lengths", [[0.8, 0.2], [0.7, 0.2, 0.1]])
@pytest.mark.parametrize("fields", [[], ["labels", "weights"]])
@pytest.mark.parametrize("random_split", [True, False])
def test_dictionary_data_module_split(lengths, fields, random_split):
    """The datamodule splits the dataset into train/val/test subsets.

    Tests that:
    * The sub-datasets have the correct lengths.
    * The dataloader returns all the fields in the DictDataset.
    """
    # Create a dataset.
    n_samples = 50
    dataset = {"data": torch.randn((n_samples, 2))}
    dataset.update({k: torch.randn((n_samples, 1)) for k in fields})
    dataset = DictDataset(dataset)

    # Splits the dataset.
    batch_size = 5
    datamodule = DictModule(
        dataset, lengths=lengths, batch_size=batch_size, random_split=random_split
    )
    datamodule.setup("fit")

    # The length of the datasets is correct.
    # The test dataloader is skipped if len(lengths) < 3.
    for length, loader_fn in zip(
        lengths,
        [
            datamodule.train_dataloader,
            datamodule.val_dataloader,
            datamodule.test_dataloader,
        ],
    ):
        loader = loader_fn()

        # Check that the dataset has the correct number of samples.
        if isinstance(length, int):
            assert len(loader.dataset) == length
        else:  # Fraction.
            assert len(loader.dataset) == int(n_samples * length)

        # Check also the fields.
        for data in loader:
            # Data is always there.
            assert data["data"].shape == (batch_size, 2)
            for field in fields:
                assert data[field].shape == (batch_size, 1)

    # An error is raised if the length of the test set has not been specified.
    if len(lengths) < 3:
        with pytest.raises(ValueError, match="you need to pass three lengths"):
            datamodule.test_dataloader()


@pytest.mark.parametrize("random_split", [True, False])
def test_dictionary_data_module_multidataset(random_split):
    """The datamodule combines multiple datasets into one."""
    n_datasets = 3
    n_samples = 10
    batch_size = 2
    lengths = [0.8, 0.2]

    # Create the datasets.
    datasets = []
    for dataset_idx in range(n_datasets):
        dataset = DictDataset(
            {
                f"data{dataset_idx}": torch.randn(n_samples, 2),
                f"labels{dataset_idx}": torch.randn(n_samples),
            }
        )
        datasets.append(dataset)

    # Create the dataloader.
    datamodule = DictModule(
        datasets, batch_size=batch_size, lengths=lengths, random_split=random_split
    )
    datamodule.setup()
    dataloader = datamodule.train_dataloader()
    assert dataloader.dataset_len == [int(lengths[0] * n_samples)] * n_datasets

    # Test that the batches are correct.
    for batch in dataloader:
        assert len(batch) == n_datasets
        for dataset_idx in range(n_datasets):
            batch_dataset = batch[f"dataset{dataset_idx}"]
            assert batch_dataset[f"data{dataset_idx}"].shape == (batch_size, 2)
            assert batch_dataset[f"labels{dataset_idx}"].shape == (batch_size,)

    # If datasets are not of the same dimension, the dataloader should explode on init.
    datasets.append(
        DictDataset(
            {
                f"data": torch.randn(n_samples + 1, 2),
                f"labels": torch.randn(n_samples + 1),
            }
        )
    )
    datamodule = DictModule(datasets, batch_size=1)
    datamodule.setup()
    with pytest.raises(ValueError, match="must have the same number of batches"):
        datamodule.train_dataloader()



#!/usr/bin/env python


# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Test objects and function in mlcolvar.cvs.unsupervised.vae.
"""


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

import os
import tempfile

import pytest
import lightning
import torch

from mlcolvar.cvs.unsupervised.vae import VariationalAutoEncoderCV
from mlcolvar.data import DictDataset, DictModule


# =============================================================================
# TESTS
# =============================================================================


@pytest.mark.parametrize("weights", [False, True])
def test_vae_cv_training(weights):
    """Run a full training of a VAECv."""
    # Create VAE CV.
    n_cvs = 2
    in_features = 8
    model = VariationalAutoEncoderCV(
        n_cvs=n_cvs,
        encoder_layers=[in_features, 6, 4],
        options={
            "norm_in": None,
            "encoder": {"activation": "relu"},
        },
    )

    # Create input data.
    batch_size = 100
    x = torch.randn(batch_size, in_features)
    data = {"data": x}

    # Create weights.
    if weights is True:
        data["weights"] = torch.rand(batch_size)

    # Train.
    datamodule = DictModule(DictDataset(data))
    trainer = lightning.Trainer(
        max_epochs=1, log_every_n_steps=2, logger=None, enable_checkpointing=False
    )
    trainer.fit(model, datamodule)

    # Eval.
    model.eval()
    x_hat = model(x)
    assert x_hat.shape == (batch_size, n_cvs)

    # Test export to torchscript.
    # This try-finally clause is a workaround for windows not allowing opening temp files twice.
    try:
        tmp_file = tempfile.NamedTemporaryFile("wb", suffix=".ptc", delete=False)
        tmp_file.close()
        model.to_torchscript(file_path=tmp_file.name, method="trace")
        model_loaded = torch.jit.load(tmp_file.name)
    finally:
        os.unlink(tmp_file.name)
    x_hat2 = model_loaded(x)
    assert torch.allclose(x_hat, x_hat2)



#!/usr/bin/env python


# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Test objects and functions in mlcolvar.cvs.multitask.multitask.
"""


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

import os
import tempfile

import pytest
import lightning
import torch

from mlcolvar.core.loss import TDALoss, FisherDiscriminantLoss, AutocorrelationLoss
from mlcolvar.cvs.cv import BaseCV
from mlcolvar.cvs.multitask.multitask import MultiTaskCV
from mlcolvar.cvs.timelagged import DeepTICA
from mlcolvar.cvs.unsupervised import AutoEncoderCV, VariationalAutoEncoderCV
from mlcolvar.data import DictDataset, DictModule


# =============================================================================
# GLOBAL VARIABLES
# =============================================================================

N_DESCRIPTORS = 4
N_CVS = 2
N_STATES = 3


# =============================================================================
# TEST UTILITY FUNCTIONS/CLASSES
# =============================================================================


class MockAuxLoss(torch.nn.Module):
    """Mock auxiliary loss for mock testing."""

    def __init__(self, in_features=N_DESCRIPTORS, out_features=N_CVS):
        super().__init__()
        self.task_specific_nn = torch.nn.Linear(in_features, out_features)

        # Store initial parameters to check that training really happens.
        self.initial_nn_weight = self.task_specific_nn.weight.detach().clone()
        self.kwargs = None

    def forward(self, data, data_lag=None, **kwargs):
        self.kwargs = kwargs
        return self.task_specific_nn(data).sum()


class MockCV(BaseCV, lightning.LightningModule):
    """Mock CV for mock testing."""

    BLOCKS = []

    def __init__(self, in_features=N_DESCRIPTORS, out_features=N_CVS):
        """Constructor."""
        super().__init__(in_features=in_features, out_features=out_features)
        self.loss_fn = MockAuxLoss(in_features, out_features)

    def training_step(self, train_batch, batch_idx):
        """Training step."""
        return self.loss_fn(**train_batch)


def create_dataset(
    dataset_type,
    weights=False,
    n_samples=40,
    n_descriptors=N_DESCRIPTORS,
    n_labels=N_STATES,
):
    """Create one of three types of datasets with random data for testing.

    dataset_type can be 'unsupervised', 'supervised', or 'time-lagged', which will
    determine the fields of the dataset:

    * 'unsupervised': 'data', ('weights')
    * 'supervised': 'data', 'labels', ('weights')
    * 'time-lagged': 'data', 'data_lag', ('weights', 'weights_lag')

    Weights are added only if ``weights`` is ``True``.

    """
    assert dataset_type in set(["supervised", "unsupervised", "time-lagged"])

    data = {"data": torch.randn((n_samples, n_descriptors))}

    # Add weights.
    if weights:
        data["weights"] = torch.rand(n_samples)

    # Data-type-specific fields.
    if dataset_type == "supervised":
        # With sequential sampling, this make sure that all labels are represented
        # in the validation and training set so that LDA/TDA don't complain.
        labels = torch.arange(n_labels, dtype=torch.get_default_dtype())
        data["labels"] = labels.repeat(n_samples // n_labels + 1)[:n_samples]
    elif dataset_type == "time-lagged":
        data["data_lag"] = torch.randn((n_samples, n_descriptors))
        if weights:
            data["weights_lag"] = torch.rand(n_samples)

    return DictDataset(data)


def create_cv(cv_name, n_descriptors=N_DESCRIPTORS, n_cvs=N_CVS):
    """Return a new CV and its associated dataset type (see create_dataset).

    cv_name can be one of 'ae', 'vae', or 'deeptica'.
    """
    if cv_name == "ae":
        returned = "unsupervised", AutoEncoderCV(
            encoder_layers=[n_descriptors, 10, n_cvs]
        )
    elif cv_name == "vae":
        returned = "unsupervised", VariationalAutoEncoderCV(
            n_cvs=n_cvs, encoder_layers=[n_descriptors, 10]
        )
    elif cv_name == "deeptica":
        returned = "time-lagged", DeepTICA(layers=[n_descriptors, 10, n_cvs])
    else:
        raise ValueError("Unrecognized cv_name.")

    # With multiple dataset, Normalization's parameters must be initialized manually.
    # This work because by default mean and range are assigned 0 and 1 value.
    returned[1].norm_in.is_initialized = True

    return returned


def create_loss(loss_name, n_states=N_STATES, n_cvs=N_CVS):
    """Return a loss and its associated dataset type (see create_dataset).

    loss_name can be one of 'lda', 'tda', and 'autocorrelation' or a list of these
    elements.

    """
    if loss_name == "autocorrelation":
        return "time-lagged", AutocorrelationLoss()
    elif loss_name == "lda":
        return "supervised", FisherDiscriminantLoss(n_states=n_states)
    elif loss_name == "tda":
        cv = TDALoss(
            n_states=N_STATES,
            target_centers=torch.linspace(-1, 1, N_STATES)
            .unsqueeze(-1)
            .repeat(1, N_CVS),
            target_sigmas=torch.tensor([0.2] * N_STATES).unsqueeze(-1).repeat(1, N_CVS),
        )
        return "supervised", cv
    else:
        raise ValueError("Unrecognized loss_name")


def create_multitask_cv_and_datasets(
    main_cv_name, weights, auxiliary_loss_names, loss_coefficients
):
    """Return a new MultiTaskCV object and a list of compatible datasets.

    main_cv_name can take all values supported in create_cv().
    weights are passed to create_dataset()
    auxiliary_loss_names can take all values supported in create_loss().
    loss_coefficients are passed directly to MultiTaskCV.__init__().

    Return the MultiTaskCV object and a list of datasets.

    """
    # Instantiate the base CV model and its dataset.
    dataset_type, main_cv = create_cv(main_cv_name)
    datasets = [create_dataset(dataset_type, weights=weights)]

    # Instantiate the auxiliary losses and its datasets.
    aux_losses = []
    for aux_loss_name in auxiliary_loss_names:
        dataset_type, loss = create_loss(aux_loss_name)
        aux_losses.append(loss)
        datasets.append(create_dataset(dataset_type))

    # Create the MultiTask CV.
    multi_cv = MultiTaskCV(main_cv, aux_losses, loss_coefficients)

    return multi_cv, datasets


# =============================================================================
# TESTS
# =============================================================================


@pytest.mark.parametrize(
    "dataset_types,weights,loss_coefficients",
    [
        (["supervised", "unsupervised"], [False, False], None),
        (["unsupervised", "supervised", "time-lagged"], [True, False, True], None),
        (["time-lagged", "unsupervised", "supervised"], [True, True, True], [2.0, 5.0]),
    ],
)
def test_multitask_loss(dataset_types, weights, loss_coefficients):
    """Auxiliary loss functions are called correctly.

    Test:
    * Auxuliary functions are all called during training.
    * Task specific layers within auxiliary losses are recognized by PyTorch and trained.
    * All the dataset keywords are passed to the losses.
    """
    # Create mock MultitaskCV.
    main_cv = MockCV()
    aux_loss_fns = [MockAuxLoss() for _ in range(len(dataset_types) - 1)]
    datasets = [create_dataset(dt, weights=w) for dt, w in zip(dataset_types, weights)]
    multi_cv = MultiTaskCV(main_cv, aux_loss_fns, loss_coefficients)

    # Do a few steps of training.
    datamodule = DictModule(datasets, shuffle=False, random_split=False)
    trainer = lightning.Trainer(
        max_epochs=2, log_every_n_steps=5, logger=None, enable_checkpointing=False
    )
    trainer.fit(multi_cv, datamodule)

    # Check that all mock loss functions have been called and that
    # the task-specific layers have been regularly trained.
    all_losses = [multi_cv.loss_fn] + [l for l in multi_cv.auxiliary_loss_fns]
    for loss in all_losses:
        assert not torch.allclose(loss.initial_nn_weight, loss.task_specific_nn.weight)

    # Check that all fields were passed.
    for loss, dataset, dataset_type, weighted in zip(
        all_losses, datasets, dataset_types, weights
    ):
        if dataset_type == "supervised":
            expected_kwargs = {"labels"}
        else:
            expected_kwargs = set()
        if weighted:
            expected_kwargs.add("weights")
            if dataset_type == "time-lagged":
                expected_kwargs.add("weights_lag")
        assert set(loss.kwargs.keys()) == expected_kwargs


@pytest.mark.parametrize(
    "main_cv_name,weights",
    [
        ("ae", False),
        ("ae", True),
        ("vae", False),
        ("vae", True),
        ("deeptica", True),  # DeepTICA currently doesn't support unweighted data.
    ],
)
@pytest.mark.parametrize(
    "auxiliary_loss_names,loss_coefficients",
    [
        (["tda"], None),
        (["lda"], None),
        (
            ["autocorrelation"],
            None,
        ),  # This ends up testing DeepTICA + autocorrelation. Doesn't make sense but we can do it.
        (["lda", "autocorrelation"], None),
        (["tda", "autocorrelation"], None),
        (["tda"], [0.5]),  # This adds a coefficient in front of the auxiliary loss
        (["lda", "autocorrelation"], [2.0, 0.2]),
    ],
)
def test_multitask_training(
    main_cv_name, weights, auxiliary_loss_names, loss_coefficients
):
    """Run a full training of a MultiTaskCV.

    Test:
    * The CV is compatible with the PyTorch Lightning's Trainer.
    * Export works correctly.
    """
    # Create the MultiTaskCV and the list of datasets.
    multi_cv, datasets = create_multitask_cv_and_datasets(
        main_cv_name, weights, auxiliary_loss_names, loss_coefficients
    )

    # Train.
    datamodule = DictModule(datasets, shuffle=False, random_split=False)
    trainer = lightning.Trainer(
        max_epochs=1, log_every_n_steps=2, logger=None, enable_checkpointing=False
    )
    trainer.fit(multi_cv, datamodule)

    # Eval.
    multi_cv.eval()
    x = datasets[0]["data"]
    x_hat = multi_cv(x)
    assert x_hat.shape == (x.shape[0], N_CVS)

    # Do round-trip through torchscript.
    # This try-finally clause is a workaround for windows not allowing opening temp files twice.
    try:
        tmp_file = tempfile.NamedTemporaryFile("wb", suffix=".ptc", delete=False)
        tmp_file.close()
        multi_cv.to_torchscript(file_path=tmp_file.name, method="trace")
        multi_cv_loaded = torch.jit.load(tmp_file.name)
    finally:
        os.unlink(tmp_file.name)
    x_hat2 = multi_cv_loaded(x)
    assert torch.allclose(x_hat, x_hat2)



import torch
import lightning
from mlcolvar.cvs import BaseCV
from mlcolvar.core import FeedForward, Normalization
from mlcolvar.core.loss import TDALoss
from mlcolvar.data import DictModule

__all__ = ["DeepTDA"]


class DeepTDA(BaseCV, lightning.LightningModule):
    """
    Deep Targeted Discriminant Analysis (Deep-TDA) CV.
    Combine the inputs with a neural-network and optimize it in a way such that
    the data are distributed accordingly to a mixture of Gaussians. The method is described in [1]_.

    **Data**: for training it requires a DictDataset with the keys 'data' and 'labels'.

    **Loss**: distance of the samples of each class from a set of Gaussians (TDALoss)

    References
    ----------
    .. [1] E. Trizio and M. Parrinello, "From enhanced sampling to reaction profiles",
        The Journal of Physical Chemistry Letters 12, 8621– 8626 (2021).

    See also
    --------
    mlcolvar.core.loss.TDALoss
        Distance from a simple Gaussian target distribution.
    """

    BLOCKS = ["norm_in", "nn"]

    # TODO n_states optional?
    def __init__(
        self,
        n_states: int,
        n_cvs: int,
        target_centers: list,
        target_sigmas: list,
        layers: list,
        options: dict = None,
        **kwargs,
    ):
        """
        Define Deep Targeted Discriminant Analysis (Deep-TDA) CV composed by a neural network module.
        By default a module standardizing the inputs is also used.

        Parameters
        ----------
        n_states : int
            Number of states for the training
        n_cvs : int
            Number of collective variables to be trained
        target_centers : list
            Centers of the Gaussian targets
        target_sigmas : list
            Standard deviations of the Gaussian targets
        layers : list
            Number of neurons per layer
        options : dict[str, Any], optional
            Options for the building blocks of the model, by default {}.
            Available blocks: ['norm_in', 'nn'].
            Set 'block_name' = None or False to turn off that block
        """

        super().__init__(in_features=layers[0], out_features=layers[-1], **kwargs)

        # =======   LOSS  =======
        self.loss_fn = TDALoss(
            n_states=n_states,
            target_centers=target_centers,
            target_sigmas=target_sigmas,
        )

        # ======= OPTIONS =======
        # parse and sanitize
        options = self.parse_options(options)
        # Save n_states
        self.n_states = n_states
        if self.out_features != n_cvs:
            raise ValueError(
                "Number of neurons of last layer should match the number of CVs!"
            )

        # check size  and type of targets
        if not isinstance(target_centers, torch.Tensor):
            target_centers = torch.Tensor(target_centers)
        if not isinstance(target_sigmas, torch.Tensor):
            target_sigmas = torch.Tensor(target_sigmas)

        if target_centers.shape != target_sigmas.shape:
            raise ValueError(
                f"Size of target_centers and target_sigmas should be the same!"
            )
        if n_states != target_centers.shape[0]:
            raise ValueError(
                f"Size of target_centers at dimension 0 should match the number of states! Expected {n_states} found {target_centers.shape[0]}"
            )
        if len(target_centers.shape) == 2:
            if n_cvs != target_centers.shape[1]:
                raise ValueError(
                    (
                        f"Size of target_centers at dimension 1 should match the number of cvs! Expected {n_cvs} found {target_centers.shape[1]}"
                    )
                )

        # ======= BLOCKS =======

        # Initialize norm_in
        o = "norm_in"
        if (options[o] is not False) and (options[o] is not None):
            self.norm_in = Normalization(self.in_features, **options[o])

        # initialize NN
        o = "nn"
        self.nn = FeedForward(layers, **options[o])

    def training_step(self, train_batch, batch_idx):
        """Compute and return the training loss and record metrics."""
        # =================get data===================
        x = train_batch["data"]
        labels = train_batch["labels"]
        # =================forward====================
        z = self.forward_cv(x)
        # ===================loss=====================
        loss, loss_centers, loss_sigmas = self.loss_fn(
            z, labels, return_loss_terms=True
        )
        # ====================log=====================+
        name = "train" if self.training else "valid"
        self.log(f"{name}_loss", loss, on_epoch=True)
        self.log(f"{name}_loss_centers", loss_centers, on_epoch=True)
        self.log(f"{name}_loss_sigmas", loss_sigmas, on_epoch=True)
        return loss


# TODO signature of tests?
import numpy as np


def test_deeptda_cv():
    from mlcolvar.data import DictDataset

    for states_and_cvs in [[2, 1], [3, 1], [3, 2], [5, 4]]:
        # get the number of states and cvs for the test run
        n_states = states_and_cvs[0]
        n_cvs = states_and_cvs[1]

        in_features, out_features = 2, n_cvs
        layers = [in_features, 4, 2, out_features]
        target_centers = np.random.randn(n_states, n_cvs)
        target_sigmas = np.random.randn(n_states, n_cvs)

        # test initialize via dictionary
        options = {"nn": {"activation": "relu"}}

        model = DeepTDA(
            n_states=n_states,
            n_cvs=n_cvs,
            target_centers=target_centers,
            target_sigmas=target_sigmas,
            layers=layers,
            options=options,
        )

        print("----------")
        print(model)

        # create dataset
        samples = 100
        X = torch.randn((samples * n_states, 2))

        # create labels
        y = torch.zeros(X.shape[0])
        for i in range(1, n_states):
            y[samples * i :] += 1

        dataset = DictDataset({"data": X, "labels": y})
        datamodule = DictModule(dataset, lengths=[0.75, 0.2, 0.05], batch_size=samples)
        # train model
        trainer = lightning.Trainer(
            accelerator="cpu", max_epochs=2, logger=None, enable_checkpointing=False
        )
        trainer.fit(model, datamodule)

        # trace model
        traced_model = model.to_torchscript(
            file_path=None, method="trace", example_inputs=X[0]
        )
        model.eval()
        assert torch.allclose(model(X), traced_model(X))


if __name__ == "__main__":
    test_deeptda_cv()



import torch
import numpy as np
from mlcolvar.core.transform.utils import Statistics
from torch.utils.data import Dataset

__all__ = ["DictDataset"]


class DictDataset(Dataset):
    """Define a torch dataset from a dictionary of lists/array/tensors and names.

    E.g. { 'data' : torch.Tensor([1,2,3,4]),
           'labels' : [0,0,1,1],
           'weights' : np.asarray([0.5,1.5,1.5,0.5]) }
    """

    def __init__(self, dictionary: dict = None, **kwargs):
        """Create a Dataset from a dictionary or from a list of kwargs.

        Parameters
        ----------
        dictionary : dict
            Dictionary with names and tensors

        """
        # assert type dict
        if (dictionary is not None) and (not isinstance(dictionary, dict)):
            raise TypeError(
                f"DictDataset requires a dictionary , not {type(dictionary)}."
            )

        # Add kwargs to dict
        if dictionary is None:
            dictionary = {}
        dictionary = {**dictionary, **kwargs}
        if len(dictionary) == 0:
            raise ValueError("Empty datasets are not supported")

        # convert to torch.Tensors
        for key, val in dictionary.items():
            if not isinstance(val, torch.Tensor):
                dictionary[key] = torch.Tensor(val)

        # save dictionary
        self._dictionary = dictionary

        # check that all elements of dict have same length
        it = iter(dictionary.values())
        self.length = len(next(it))
        if not all(len(l) == self.length for l in it):
            raise ValueError("not all arrays in dictionary have same length!")

    def __getitem__(self, index):
        if isinstance(index, str):
            # raise TypeError(f'Index ("{index}") should be a slice, and not a string. To access the stored dictionary use .dictionary["{index}"] instead.')
            return self._dictionary[index]
        else:
            slice_dict = {}
            for key, val in self._dictionary.items():
                slice_dict[key] = val[index]
            return slice_dict

    def __setitem__(self, index, value):
        if isinstance(index, str):
            # check lengths
            if len(value) != len(self):
                raise ValueError(
                    f"length of value ({len(value)}) != length of dataset ({len(self)})."
                )
            self._dictionary[index] = value
        else:
            raise NotImplementedError(
                f"Only string indexes can be set, {type(index)} is not supported."
            )

    def __len__(self):
        value = next(iter(self._dictionary.values()))
        return len(value)

    def get_stats(self):
        """Compute statistics ('mean','Std','Min','Max') of the dataset.

        Returns
        -------
        stats
            dictionary of dictionaries with statistics
        """
        stats = {}
        for k in self.keys:
            stats[k] = Statistics(self._dictionary[k]).to_dict()
        return stats

    def __repr__(self) -> str:
        string = "DictDataset("
        for key, val in self._dictionary.items():
            string += f' "{key}": {list(val.shape)},'
        string = string[:-1] + " )"
        return string

    @property
    def keys(self):
        return tuple(self._dictionary.keys())


def test_DictDataset():
    # from list
    dataset_dict = {
        "data": torch.Tensor([[1.0], [2.0], [0.3], [0.4]]),
        "labels": [0, 0, 1, 1],
        "weights": np.asarray([0.5, 1.5, 1.5, 0.5]),
    }

    dataset = DictDataset(dataset_dict)
    print(len(dataset))
    print(dataset[0])
    print(dataset[0:2]["data"])
    print(dataset[0:2]["data"].dtype)

    # test with dataloader
    from torch.utils.data import DataLoader

    loader = DataLoader(dataset, batch_size=1)
    batch = next(iter(loader))
    print(batch["data"])

    # test with fastdataloader
    from .dataloader import DictLoader

    loader = DictLoader(dataset, batch_size=1)
    batch = next(iter(loader))
    print(batch)


if __name__ == "__main__":
    test_DictDataset()



import torch
import numpy as np
from bisect import bisect_left
from mlcolvar.data import DictDataset
import warnings

# optional packages
# pandas
try:
    import pandas as pd

    PANDAS_IS_INSTALLED = True
except ImportError:
    PANDAS_IS_INSTALLED = False
# tqdm (progress bar)
try:
    from tqdm import tqdm

    TQDM_IS_INSTALLED = True
except ImportError:
    TQDM_IS_INSTALLED = False

__all__ = ["find_timelagged_configurations", "create_timelagged_dataset"]


def closest_idx(array, value):
    """
    Find index of the element of 'array' which is closest to 'value'.
    The array is first converted to a np.array in case of a tensor.
    Note: it does always round to the lowest one.

    Parameters:
        array (tensor/np.array)
        value (float)

    Returns:
        pos (int): index of the closest value in array
    """
    if type(array) is np.ndarray:
        pos = bisect_left(array, value)
    else:
        pos = bisect_left(array.numpy(), value)
    if pos == 0:
        return 0
    elif pos == len(array):
        return -1
    else:
        return pos - 1


# evaluation of tprime from simulation time and logweights
def tprime_evaluation(t, logweights=None):
    """
    Estimate the accelerated time if a set of (log)weights is given

    Parameters
    ----------
    t : array-like,
        unbias time series,
    logweights : array-like,optional
        logweights to evaluate rescaled time as dt' = dt*exp(logweights)
    """

    # rescale time with log-weights if given
    if logweights is not None:
        # compute time increment in simulation time t
        dt = np.round(t[1] - t[0], 5)
        # sanitize logweights
        logweights = torch.Tensor(logweights)
        # when the bias is not deposited the value of bias potential is minimum
        logweights -= torch.max(logweights)
        # note: exp(logweights/lognorm) != exp(logweights)/norm, where norm is sum_i beta V_i
        """ possibilities:
            1) logweights /= torch.min(logweights) -> logweights belong to [0,1] 
            2) pass beta as an argument, then logweights *= beta
            3) tprime = dt * torch.cumsum( torch.exp( torch.logsumexp(logweights,0) ) ,0)
            4) tprime = dt *torch.exp ( torch.log (torch.cumsum (torch.exp(logweights) ) ) )
        """
        lognorm = torch.logsumexp(logweights, 0)
        logweights /= lognorm
        # compute instantaneus time increment in rescaled time t'
        d_tprime = torch.exp(logweights) * dt
        # calculate cumulative time t'
        tprime = torch.cumsum(d_tprime, 0)
    else:
        tprime = t

    return tprime


def find_timelagged_configurations(
    x: torch.Tensor,
    t: torch.Tensor,
    lag_time: float,
    logweights: torch.Tensor = None,
    progress_bar: bool = True,
):
    """Searches for all the pairs which are distant 'lag' in time, and returns the weights associated.
    If logweights are provided they will be returned both for x_t and x_t+lag (used only for `reweight_mode=weights_t` of create_time_lagged_dataset).

    Parameters
    ----------
    x : torch.Tensor
        array whose columns are the descriptors and rows the time evolution
    t : torch.Tensor
        array with the simulation time
    lag_time : float
        lag-time
    logweights : torch.Tensor, optional
        logweights to be returned
    progress_bar : bool, optional
        display progress bar with tqdm (if installed), by default True

    Returns
    -------
    x_t: torch.Tensor
        descriptors at time t
    x_lag: torch.Tensor
        descriptors at time t+lag
    w_t: torch.Tensor
        weights at time t
    w_lag: torch.Tensor
        weights at time t+lag
    """

    # define lists
    x_t = []
    x_lag = []
    w_t = []
    w_lag = []
    # find maximum time idx
    idx_end = closest_idx(t, t[-1] - lag_time)
    # start_j = 0
    N = len(t)

    def progress(iter, progress_bar=progress_bar):
        if progress_bar and TQDM_IS_INSTALLED:
            return tqdm(iter)
        else:
            warnings.warn(
                "Monitoring the progress for the search of time-lagged configurations with a progress_bar requires `tqdm`."
            )
            return iter

    # sanitize logweights if given
    calculate_weights = True
    if logweights is not None:
        calculate_weights = False
        if len(logweights) != len(x):
            raise ValueError(
                f"Length of logweights ({len(logweights)}) is different from length of data ({len(x)})."
            )
        logweights = torch.Tensor(logweights)
        weights = torch.exp(logweights)

    # loop over time array and find pairs which are far away by lag_time
    for i in progress(range(idx_end)):
        stop_condition = lag_time + t[i + 1]
        n_j = 0

        for j in range(i, N):
            if (t[j] < stop_condition) and (t[j + 1] > t[i] + lag_time):
                x_t.append(x[i])
                x_lag.append(x[j])
                deltaTau = min(t[i + 1] + lag_time, t[j + 1]) - max(
                    t[i] + lag_time, t[j]
                )

                if calculate_weights:
                    w_lag.append(deltaTau)
                else:
                    w_lag.append(weights[i])
                # if n_j == 0: #assign j as the starting point for the next loop
                #    start_j = j
                n_j += 1
            elif t[j] > stop_condition:
                break
        for k in range(n_j):
            if calculate_weights:
                w_t.append((t[i + 1] - t[i]) / float(n_j))
            else:
                if n_j > 1:
                    print(n_j)
                w_t.append(weights[i] / float(n_j))

    x_t = torch.stack(x_t) if type(x) == torch.Tensor else torch.Tensor(x_t)
    x_lag = torch.stack(x_lag) if type(x) == torch.Tensor else torch.Tensor(x_lag)

    w_t = torch.Tensor(w_t)
    w_lag = torch.Tensor(w_lag)

    return x_t, x_lag, w_t, w_lag


def create_timelagged_dataset(
    X: torch.Tensor,
    t: torch.Tensor = None,
    lag_time: float = 1,
    reweight_mode: str = None,
    logweights: torch.Tensor = None,
    tprime: torch.Tensor = None,
    interval: list = None,
    progress_bar: bool = False,
):
    """
    Create a DictDataset of time-lagged configurations.

    In case of biased simulations the reweight can be performed in two different ways (``reweight_mode``):

    1. ``rescale_time`` : the search for time-lagged pairs is performed in the accelerated time (dt' = dt*exp(logweights)), as described in [1]_ .
    2. ``weights_t`` : the weight of each pair of configurations (t,t+lag_time) depends only on time t (logweights(t)), as done in [2]_ , [3]_ .

    If reweighting is None and tprime is given the `rescale_time` mode is used. If instead only the logweights are specified the user needs to choose the reweighting mode.

    References
    ----------
    .. [1] Y. I. Yang and M. Parrinello, “Refining collective coordinates and improving free energy
        representation in variational enhanced sampling,” JCTC 14, 2889–2894 (2018).
    .. [2] J. McCarty and M. Parrinello, "A variational conformational dynamics approach to the selection
        of collective variables in meta- dynamics,” JCP 147, 204109 (2017).
    .. [3] H. Wu, et al. "Variational Koopman models: Slow collective variables and molecular kinetics
        from short off-equilibrium simulations." JCP 146.15 (2017).

    Parameters
    ----------
    X : array-like
        input descriptors
    t : array-like, optional
        time series, by default np.arange(len(X))
    reweight_mode: str, optional
        how to do the reweighting, see documentation, by default none
    lag_time: float, optional
        lag between configurations, by default = 10
    logweights : array-like,optional
        logweight of each configuration (typically beta*bias)
    tprime : array-like,optional
        rescaled time estimated from the simulation. If not given and `reweighting_mode`=`rescale_time` then `tprime_evaluation(t,logweights)` is used
    interval : list or np.array or tuple, optional
        Range for slicing the returned dataset. Useful to work with batches of same sizes. Recall that with different lag_times one obtains different datasets, with different lengths
    progress_bar: bool
        Display progress bar with tqdm

    Returns
    -------
    dataset: DictDataset
        Dataset with keys 'data', 'data_lag' (data at time t and t+lag), 'weights', 'weights_lag' (weights at time t and t+lag).

    """

    if PANDAS_IS_INSTALLED:
        # check if dataframe
        if type(X) == pd.core.frame.DataFrame:
            X = X.values
        if type(t) == pd.core.frame.DataFrame:
            t = t.values

    # check reweigthing mode if logweights are given:
    # 1) if rescaled time tprime is given
    if tprime is not None:
        if reweight_mode is None:
            reweight_mode = "rescale_time"
        elif reweight_mode != "rescale_time":
            raise ValueError(
                "The `reweighting_mode` needs to be equal to `rescale_time`, and not {reweight_mode} if the rescale time `tprime` is given."
            )
    # 2) if logweights are given
    elif logweights is not None:
        if reweight_mode is None:
            reweight_mode = "rescale_time"
            # TODO output warning or error if mode not specified?
            # warnings.warn('`reweight_mode` not specified, setting it to `rescale_time`.')

    # define time if not given
    if t is None:
        t = torch.arange(0, len(X))
    else:
        if len(t) != len(X):
            raise ValueError(
                f"The length of t ({len(t)}) is different from the one of X ({len(X)}) "
            )

    # define tprime if not given:
    if reweight_mode == "rescale_time":
        if tprime is None:
            tprime = tprime_evaluation(t, logweights)
    else:
        tprime = t

    # find pairs of configurations separated by lag_time
    x_t, x_lag, w_t, w_lag = find_timelagged_configurations(
        X,
        tprime,
        lag_time=lag_time,
        logweights=logweights if reweight_mode == "weights_t" else None,
        progress_bar=progress_bar,
    )

    # return only a slice of the data (N. Pedrani)
    if interval is not None:
        # convert to a list
        data = list(x_t, x_lag, w_t, w_lag)
        # assert dimension of interval
        assert len(interval) == 2
        # modifies the content of data by slicing
        for i in range(len(data)):
            data[i] = data[i][interval[0] : interval[1]]
        x_t, x_lag, w_t, w_lag = data

    dataset = DictDataset(
        {"data": x_t, "data_lag": x_lag, "weights": w_t, "weights_lag": w_lag}
    )

    return dataset


def test_create_timelagged_dataset():
    in_features = 2
    n_points = 100
    X = torch.rand(n_points, in_features) * 100

    # unbiased case
    t = np.arange(n_points)
    dataset = create_timelagged_dataset(X, t, lag_time=10)
    print(len(dataset))

    # reweight mode rescale_time (default)
    logweights = np.random.rand(n_points)
    dataset = create_timelagged_dataset(X, t, logweights=logweights)
    print(len(dataset))

    # reweight mode weights_t
    logweights = np.random.rand(n_points)
    dataset = create_timelagged_dataset(
        X, t, logweights=logweights, reweight_mode="weights_t"
    )
    print(len(dataset))


if __name__ == "__main__":
    test_create_timelagged_dataset()



__all__ = [
    "BaseCV",
    "DeepLDA",
    "DeepTICA",
    "DeepTDA",
    "AutoEncoderCV",
    "RegressionCV",
    "MultiTaskCV",
]

from .cv import BaseCV
from .unsupervised import *
from .supervised import *
from .timelagged import *
from .multitask import *



import torch
from torch import Tensor
from typing import Optional


def generalized_eigh(A: Tensor, B: Tensor) -> tuple:
    r"""A workaround to solve a real symmetric generalized eigenvalue problem :math:`Av = \lambda Bv` using the eigenvalue decomposition of :math:`B^{-1/2}AB^{-1/2}`. This method is not numerically efficient.

    Parameters
    ----------
    A : Tensor

    B : Tensor

    Returns
    -------
    tuple
        Eigenvalues and eigenvectors of the generalized eigenvalue problem.
    """
    Lambda, Q = torch.linalg.eigh(B)
    rsqrt_Lambda = torch.diag(Lambda.rsqrt())
    rsqrt_B = Q @ rsqrt_Lambda
    _A = 0.5 * (
        rsqrt_B.T @ (A @ rsqrt_B) + rsqrt_B.T @ ((A.T) @ rsqrt_B)
    )  # Force Symmetrization
    values, _tmp_vecs = torch.linalg.eigh(_A)
    vectors = rsqrt_B @ _tmp_vecs
    return values, vectors


def spd_norm(vecs: Tensor, spd_matrix: Tensor) -> Tensor:
    """Compute the norm of a set of vectors with respect to a symmetric positive definite matrix.

    Parameters
    ----------
    vecs : Tensor
        Two dimensional tensor whose columns are the vectors whose norm is to be computed.
    spd_matrix : Tensor
        Symmetric positive matrix. Warning: this matrix is not checked for symmetry or positive definiteness.

    Returns
    -------
    Tensor
        One dimensional tensor whose i-th element is the norm of the i-th column of vecs with respect to spd_matrix.
    """
    _v = torch.mm(spd_matrix, vecs)
    _v_T = torch.mm(spd_matrix.T, vecs)
    return torch.sqrt(0.5 * torch.linalg.vecdot(vecs, _v + _v_T, dim=0).real)


def reduced_rank_eig(
    input_covariance: Tensor,
    lagged_covariance: Tensor,  # C_{0t}
    tikhonov_reg: float,
    rank: Optional[int] = None,
) -> tuple:
    """Reduced rank regression algorithm, as described in [1]_.

    Parameters
    ----------
    input_covariance : Tensor

    lagged_covariance : Tensor

    rank : Optional[int], optional
        Rank of the final estimator, by default None

    Returns
    -------
    tuple
        A tuple containing the eigenvalues and eigenvectors of the Koopman operator.

    References
    ----------
    .. [1] V. Kostic, P. Novelli, A. Maurer, C. Ciliberto, L. Rosasco, and M. Pontil, "Learning Dynamical Systems via Koopman Operator Regression in Reproducing Kernel Hilbert Spaces" (2022).
    """
    n = input_covariance.shape[0]
    reg_input_covariance = input_covariance + tikhonov_reg * torch.eye(
        n, dtype=input_covariance.dtype, device=input_covariance.device
    )

    _crcov = torch.mm(lagged_covariance, lagged_covariance.T)
    _, _vectors = generalized_eigh(_crcov, reg_input_covariance)

    _norms = spd_norm(_vectors, reg_input_covariance)
    vectors = _vectors * (1 / _norms)

    if rank is not None:
        _, idxs = torch.topk(vectors.values, rank)
        U = vectors[:, idxs]
    else:
        U = vectors

    # U@(U.T)@Tw = v w -> (U.T)@T@Uq = vq and w = Uq
    values, Q = torch.linalg.eig((U.T) @ (lagged_covariance @ U))
    return values, U @ Q


def cholesky_eigh(A, B, reg_B=1e-6, n_eig=None):
    """
    -- Generalized eigenvalue problem: A * v_i = lambda_i * B * v_i --

    First apply cholesky decomposition to B and then solve the generalized eigvalue problem.

    Notes
    -----
    The eigenvecs object which is returned is a matrix whose column eigvecs[:,i] is the eigenvector associated to eigvals[i]
    """

    # check that both matrices are symmetric
    if not torch.allclose(A.transpose(0, 1), A):
        raise ValueError(
            "The matrices need to be symmetric to solve the generalized eigenvalue problem via cholesky decomposition. A >> ",
            A,
        )
    if not (torch.allclose(B.transpose(0, 1), B)):
        raise ValueError(
            "The matrices need to be symmetric to solve the generalized eigenvalue problem via cholesky decomposition. A >> ",
            B,
        )

    # (0) regularize B matrix before cholesky
    if reg_B is not None:
        B = B + reg_B * torch.eye(B.shape[0]).to(B.device)

    # (1) use cholesky decomposition for B
    L = torch.linalg.cholesky(B, upper=False)

    # (2) define new matrix using cholesky decomposition
    L_t = torch.t(L)
    L_ti = torch.inverse(L_t)
    L_i = torch.inverse(L)
    A_new = torch.matmul(torch.matmul(L_i, A), L_ti)

    # (3) find eigenvalues and vectors of A_new
    eigvals, eigvecs = torch.linalg.eigh(A_new, UPLO="L")
    # sort
    eigvals, indices = torch.sort(eigvals, 0, descending=True)
    eigvecs = eigvecs[:, indices]

    # (4) return to original eigenvectors
    eigvecs = torch.matmul(L_ti, eigvecs)

    # (5) normalize them
    eigvecs = torch.nn.functional.normalize(eigvecs, dim=0)
    # set the first component positive
    eigvecs = eigvecs.mul(torch.sign(eigvecs[0, :]).unsqueeze(0).expand_as(eigvecs))

    # (6) keep only first n_eig eigvals and eigvecs
    if n_eig is not None:
        eigvals = eigvals[:n_eig]
        eigvecs = eigvecs[:, :n_eig]

    return eigvals, eigvecs


""" TODO implement also the non-symmetric version?
#Compute the pseudoinverse (Moore-Penrose inverse) of C_0. if det(C_0) != 0 then the usual inverse is computed
            C_new = torch.matmul(torch.pinverse(C_0),C_lag)
            #find eigenvalues and vectors of C_new
            #NOTE: torch.linalg.eig() returns complex tensors of dtype cfloat or cdouble
            #          rather than real tensors mimicking complex tensors. 
            #          For future developments it would be necessary to take either only the real part
            #          or only the complex part or only the magnitude of the complex eigenvectors and eigenvalues 
            eigvals, eigvecs = torch.linalg.eig(C_new, UPLO='L')
"""


def correlation_matrix(x, y, w=None, symmetrize=True):
    """Compute the correlation matrix between x and y with weights w

    Parameters
    ----------
    x : torch.Tensor
        first array
    y : torch.Tensor
        second array
    w : torch.Tensor, optional
        weights, by default None
    symmetrize: bool, optional
        whether to return 0.5 * (C + C^T), by default True

    Returns
    -------
    torch.Tensor
        correlation matrix

    """
    # TODO Add assert on shapes

    if w is None:  # TODO simplify it in the unbiased case?
        w = torch.ones(x.shape[0])
        w = w.to(x.device)

    # compute correlation matrix
    corr = torch.einsum("ij, ik, i -> jk", x, y, w)
    corr /= torch.sum(w)

    if symmetrize:
        corr = 0.5 * (corr + corr.T)

    return corr


def compute_average(x, w=None):
    """Compute (weighted) average on a batch

    Parameters
    ----------
    x : torch.Tensor
        Input data of shape
    w : torch.Tensor, optional
        Weights, by default None

    Returns
    -------
    torch.Tensor
        (weighted) mean of inputs

    """
    if w is not None:
        ave = torch.einsum("ij,i ->j", x, w) / torch.sum(w)
    else:
        ave = torch.mean(x.T, 1, keepdim=False).T

    return ave



"""Machine learning collective variables"""

__all__ = ["loss", "nn", "transform", "stats"]

# Add imports here
from .core import *
from .data import *
from .cvs import *

from ._version import __version__
